#ifndef __DRIVER_H_
#define __DRIVER_H_

struct config_item {
    char* name;
    int default_value;
    int current_value;
    int new_value;
};

/**
 * Common structure for driver definition.
 */
struct driver_common {
    /**
     * Initialize a driver
     *
     * This method is called at boot for all drivers
     * @param drv the driver to initialize
     * @param priv the data specific to the driver
     *
     * @return int error code or 0 if success
     */
    int (*init)(struct driver_common* drv, void* priv);

    /**
     * Deinitialize a driver
     *
     * This method can be called in shutdown in order to
     * put hardware in a safe power-down condition.
     *
     * @param drv the driver to deinitialize
     * @param priv the data specific to the driver
     *
     * @return int error code or 0 if success
     */
    int (*deinit)(struct driver_common* drv, void* priv);

    /**
     * Read configuration
     *
     * This method fills current value for all configuration
     * items defined for this driver
     *
     * @param drv the driver to deinitialize
     *
     * @return int error code or 0 if success
     */
    int (*read_config)(struct driver_common* drv);

    /**
     * Apply configuration
     *
     * This method applies the new values set in the configuration
     * elements and make them new values.
     *
     * @param drv the driver to deinitialize
     *
     * @return int error code or 0 if success
     */
    int (*apply_config)(struct driver_common* drv);

    struct config_item* config_items;
    int config_items_count;

    /**
     * Pointer to the driver's registry item
     */
    struct driver_item* reg;
};

/** A device driver have to be initialized at startup */
#define DRIVER_TYPE_DEVICE 1
/** An interface type driver will be initialized when
 * attached to a bus */
#define DRIVER_TYPE_INTERFACE 2

/**
 * Structure holding a driver item in the registry.
 */
struct driver_item {
    /** Name of the driver or interface */
    char* name;
    /** Class name of the driver or interface
     * This field allows to lookup driver by roles
     * for instance, all i2c bus will be of class "i2c",
     * all power monitoring will be of class "pmon"
     */
    char* class;
    /** The actual driver instance pointer */
    struct driver_common* driver;
    /** The type of driver (interface or driver instance) */
    int type;
    /** The private data passed to init in case of driver instance */
    void* priv;
};

#define DECLARE_DRIVER(_name_, _class_, _driver_, _priv_) \
    _DECLARE_DRIVER(_name_, _class_, _driver_, _priv_, DRIVER_TYPE_DEVICE)
#define DECLARE_INTERFACE(_name_, _class_, _driver_, _priv_) \
    _DECLARE_DRIVER(_name_, _class_, _driver_, _priv_, DRIVER_TYPE_INTERFACE)
/**
 * Declare a driver in the registry.
 */
#if defined(__ARMCC_VERSION) && (__ARMCC_VERSION >= 6010050)
#define _DECLARE_DRIVER(_name_, _class_, _driver_, _priv_, _type)                                  \
    struct driver_item __attribute__((section("DRV"))) __attribute__((used)) __driver_##_name_ = { \
        .name = #_name_, .class = #_class_, .driver = _driver_, .type = _type, .priv = _priv_};

#elif defined(__GNUC__)
#define _DECLARE_DRIVER(_name_, _class_, _driver_, _priv_, _type)                        \
    struct driver_item __attribute__((section(".drv." #_class_ #_name_))) __attribute__( \
        (used)) __driver_##_name_ = {                                                    \
        .name = #_name_, .class = #_class_, .driver = _driver_, .type = _type, .priv = _priv_};

#endif
/**
 * This function should be called by the runtime before any driver is used.
 * It will iterate over all drivers and call their init function.
 *
 * @return 0 if init succesfull, error code if an error occured
 */
int driver_init(void);

/**
 * Search for a driver given its name.
 *
 * @param name the name of the driver to look for
 *
 * @return struct driver_common * the pointer to the driver handler NULL if driver not found.
 */
struct driver_common* driver_lookup(char* name);

/**
 * Search for an interface given its name.
 *
 * @param name the name of the interface to look for
 *
 * @return struct driver_common * the pointer to the interface operations.
 */
struct driver_common* interface_lookup(char* name);

/**
 * This function should be called by the runtime before any driver is used.
 * It will iterate over all drivers and call their init function.
 *
 * @return 0 if init successful, error code if an error occurred
 */
int driver_set_config(struct driver_common* drv, char* name, int value);

/**
 * This function should be called by the runtime before any driver is used.
 * It will iterate over all drivers and call their init function.
 *
 * @return 0 if init successful, error code if an error occurred
 */
int driver_apply_config(struct driver_common* drv);

#endif
