#include "driver.h"
#include <stdlib.h>
#include "bsp_error.h"
#include "cli_cmd.h"
#include "custom_sections.h"
#include "i2c_bus.h"

DECLARE_CUSTOM_SECTION(drv, struct driver_item, DRV);

int driver_init() {
    struct driver_item* drv = drv_start;
    while (drv < drv_end) {
        int ret;
        if (drv->type == DRIVER_TYPE_DEVICE && drv->driver->init) {
            drv->driver->reg = drv;
            ret = drv->driver->init(drv->driver, drv->priv);
            if (drv->driver->read_config)
                drv->driver->read_config(drv->driver);
            if (ret && ret != ENOT_IMPLEMENTED) {
                return ret;
            }
        }
        drv++;
    }
    return 0;
}

struct driver_common* driver_lookup(char* name) {
    struct driver_item* drv = drv_start;
    while (drv < drv_end) {
        if (drv->type == DRIVER_TYPE_DEVICE && !strcmp(drv->name, name)) {
            if (!drv->driver->reg) {
                drv->driver->reg = drv;
            }
            return drv->driver;
        }
        if (drv->type == DRIVER_TYPE_DEVICE && !strcmp(drv->class, "i2c")) {
            struct i2c_bus_driver* b = (struct i2c_bus_driver*)drv->driver;
            struct i2c_device* d = b->devices;
            while (d) {
                if (!strcmp(d->name, name)) {
                    return &d->driver;
                }
                d = d->next;
            }
        }
        drv++;
    }
    return NULL;
}

struct driver_common* interface_lookup(char* name) {
    struct driver_item* drv = drv_start;
    while (drv < drv_end) {
        if (drv->type == DRIVER_TYPE_INTERFACE && !strcmp(drv->name, name)) {
            return drv->driver;
        }
        drv++;
    }
    return NULL;
}

int driver_apply_config(struct driver_common* drv) {
    if (drv->apply_config) {
        return drv->apply_config(drv);
    }
    return ENOT_IMPLEMENTED;
}

static void print_config(struct cli_cmd_ctx* ctx, struct config_item* config_items) {
    cli_printf(ctx, "\t%16s: %8d %8d %8d\r\n", config_items->name, config_items->current_value,
               config_items->new_value, config_items->default_value);
}

static void show_config(char* name, struct driver_common* drv, struct cli_cmd_ctx* ctx) {
    cli_printf(ctx, "driver %s:\r\n", name, drv->config_items_count, drv->config_items);
    for (int i = 0; i < drv->config_items_count; i++) {
        print_config(ctx, drv->config_items + i);
    }
}

int config_cli_handler(int argc, char** argv, struct cli_cmd_ctx* ctx) {
    int ret = SUCCESS;

    if (argc == 1) {
        struct driver_item* drv = drv_start;
        while (drv < drv_end) {
            if (drv->type == DRIVER_TYPE_DEVICE && drv->driver->config_items != NULL) {
                show_config(drv->name, drv->driver, ctx);
            }
            if (drv->type == DRIVER_TYPE_DEVICE && !strcmp(drv->class, "i2c")) {
                struct i2c_bus_driver* b = (struct i2c_bus_driver*)drv->driver;
                struct i2c_device* d = b->devices;
                while (d) {
                    if (d->driver.config_items)
                        show_config(d->name, &d->driver, ctx);
                    d = d->next;
                }
            }
            drv++;
        }
    } else if (argc >= 2) {
        struct driver_common* drv = driver_lookup(argv[1]);
        if (drv == NULL) {
            return ENOT_IMPLEMENTED;
        }
        if (drv->config_items != NULL) {
            if (argc < 3) {
                show_config(argv[1], drv, ctx);
            } else {
                if (!strcmp(argv[2], "apply")) {
                    return drv->apply_config(drv);
                } else if (argc == 3) {
                    char* item = argv[2];
                    int indx = 0;
                    for (indx = 0; indx < drv->config_items_count; indx++) {
                        if (!strcmp(drv->config_items[indx].name, item)) {
                            print_config(ctx, drv->config_items + indx);
                            break;
                        }
                    }

                    if (indx == drv->config_items_count) {
                        ret |= EFAILURE;
                        cli_printf(ctx, "config %s not found for %s\r\n", item, argv[1]);
                    }
                    return ret;
                }
                int index = 2;
                while (index < argc) {
                    char* item = argv[index];
                    int value = strtoul(argv[index + 1], NULL, 0);
                    int indx = 0;
                    for (indx = 0; indx < drv->config_items_count; indx++) {
                        if (!strcmp(drv->config_items[indx].name, item)) {
                            drv->config_items[indx].new_value = value;
                            break;
                        }
                    }
                    if (indx == drv->config_items_count) {
                        ret |= EFAILURE;
                        cli_printf(ctx, "config %s not found for %s\r\n", item, argv[1]);
                    }
                    index += 2;
                }
            }
        }
    }
    return ret;
}

DECLARE_CMD(config, config_cli_handler);
