#include <stddef.h>
#include <stdint.h>
#include "bsp_error.h"
#include "i2c/ina233.h"
#include "i2c_bus.h"
#include "power_monitor.h"

#define INA233_MFR_CALIB_REG 0xd4
#define INA233_VBUS_VOLTAGE_REG 0x88
#define INA233_INPUT_CURRENT_REG 0x89
#define INA233_INPUT_POWER_REG 0x97
#define INA233_DEVICE_CONFIG_REG 0xd5
#define INA233_CLEAR_EIN_REG 0xd6
#define INA233_TI_MFR_ID_REG 0xe0
#define INA233_TI_MFR_MODEL_REG 0xe1
#define INA233_TI_MFR_REVISION_REG 0xe2
#define INA233_MFR_ID_REG 0x99
#define INA233_MFR_MODEL_REG 0x9a
#define INA233_MFR_REVISION_REG 0x9b

/*To be read form board config file*/
#define BUS_VOLTAGE_LSB 1250
#define CURRENT_LSB 1000 /*Value for a 0.001 ohm shunt and a 1mA LSb*/
#define POWER_LSB (25 * CURRENT_LSB)

struct config_item ina233_config_items[] = {{.name = "calib", .default_value = 0}};

int ina233_lld_get_voltage(struct power_monitor_driver* drv, int* volt) {
    int ret = SUCCESS;
    uint8_t buf[2];

    struct i2c_device* d = (struct i2c_device*)drv->priv;

    ret = i2c_read_reg(d, INA233_VBUS_VOLTAGE_REG, buf, 2);

    if (0 < ret) {
        *volt = (((buf[0]) | (buf[1] << 8)) * BUS_VOLTAGE_LSB);
    }

    return ret;
}

int ina233_lld_get_current(struct power_monitor_driver* drv, int* current) {
    int ret = SUCCESS;
    uint8_t buf[2];
    struct i2c_device* d = (struct i2c_device*)drv->priv;

    ret = i2c_read_reg(d, INA233_INPUT_CURRENT_REG, buf, 2);

    if (0 < ret) {
        *current = (((buf[0]) | (buf[1] << 8)) * CURRENT_LSB);
    }
    return ret;
}

int ina233_lld_get_power(struct power_monitor_driver* drv, int* power) {
    int ret = SUCCESS;
    uint8_t buf[2];
    struct i2c_device* d = (struct i2c_device*)drv->priv;

    ret = i2c_read_reg(d, INA233_INPUT_POWER_REG, buf, 2);

    if (0 < ret) {
        *power = (((buf[0]) | (buf[1] << 8)) * POWER_LSB);
    }

    return ret;
}

int ina233_lld_init(struct driver_common* drv, void* priv) {
    (void)drv;

    uint8_t buf[2];

    ((struct power_monitor_driver*)drv)->priv = priv;
    struct i2c_device* i2c_dev = (struct i2c_device*)priv;

    struct ina233_calib_data* calib = (struct ina233_calib_data*)i2c_dev->priv;

    /*LSB First*/
    buf[1] = ((calib->mfr_calib_reg >> 8) & 0xFF);
    buf[0] = (calib->mfr_calib_reg & 0xFF);

    return i2c_write_reg(i2c_dev, INA233_MFR_CALIB_REG, buf, 2);
}

int ina233_lld_deinit(struct driver_common* drv, void* priv) {
    (void)drv;
    (void)priv;

    return ENOT_IMPLEMENTED;
}

struct power_monitor_ops ina233_lld_ops = {
    .read_voltage = ina233_lld_get_voltage,
    .read_current = ina233_lld_get_current,
    .read_power = ina233_lld_get_power,
};

int ina233_read_config(struct driver_common* drv) {
    struct i2c_device* i2c_dev = (struct i2c_device*)((struct power_monitor_driver*)drv)->priv;
    struct ina233_calib_data* calib = (struct ina233_calib_data*)i2c_dev->priv;

    drv->config_items[0].current_value = calib->mfr_calib_reg;

    return SUCCESS;
}

int ina233_apply_config(struct driver_common* drv) {
    struct i2c_device* i2c_dev = (struct i2c_device*)((struct power_monitor_driver*)drv)->priv;
    uint8_t buf[2];
    drv->config_items[0].current_value = drv->config_items[0].new_value;
    buf[1] = ((drv->config_items[0].current_value >> 8) & 0xFF);
    buf[0] = (drv->config_items[0].current_value & 0xFF);

    return i2c_write_reg(i2c_dev, INA233_MFR_CALIB_REG, buf, 2);
}

struct {
    struct driver_common driver;
    struct power_monitor_ops* ops;
    void* priv;
} ina233_lld_drv = {
    .driver.init = ina233_lld_init,
    .driver.deinit = ina233_lld_deinit,
    .driver.read_config = ina233_read_config,
    .driver.apply_config = ina233_apply_config,
    .driver.config_items = ina233_config_items,
    .driver.config_items_count = sizeof(ina233_config_items) / sizeof(ina233_config_items[0]),
    .ops = &ina233_lld_ops,
};

DECLARE_INTERFACE(ina233, pmon, &ina233_lld_drv.driver, NULL);
